@@ -1,16 +1,16 @@
-{-# LANGUAGE BangPatterns, CPP, FlexibleInstances, GADTs, OverloadedStrings,
+{-# LANGUAGE BangPatterns, FlexibleInstances, GADTs, OverloadedStrings,
     Rank2Types, RecordWildCards, TypeFamilies, TypeSynonymInstances #-}
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 -- |
 -- Module      :  Data.Attoparsec.Text.Internal
--- Copyright   :  Bryan O'Sullivan 2011
+-- Copyright   :  Bryan O'Sullivan 2007-2015
 -- License     :  BSD3
 --
 -- Maintainer  :  bos@serpentine.com
 -- Stability   :  experimental
 -- Portability :  unknown
 --
--- Simple, efficient parser combinators for 'T.Text' strings, loosely
+-- Simple, efficient parser combinators for 'Text' strings, loosely
 -- based on the Parsec library.
 
 module Data.Attoparsec.Text.Internal
@@ -24,8 +24,6 @@ module Data.Attoparsec.Text.Internal
     , parseOnly
 
     -- * Combinators
-    , (<?>)
-    , try
     , module Data.Attoparsec.Combinator
 
     -- * Parsing individual characters
@@ -35,7 +33,10 @@ module Data.Attoparsec.Text.Internal
     , skip
     , char
     , notChar
+
+    -- ** Lookahead
     , peekChar
+    , peekChar'
 
     -- ** Character classes
     , inClass
@@ -48,6 +49,7 @@ module Data.Attoparsec.Text.Internal
     , asciiCI
     , take
     , scan
+    , runScanner
     , takeWhile
     , takeWhile1
     , takeTill
@@ -56,117 +58,39 @@ module Data.Attoparsec.Text.Internal
     , takeText
     , takeLazyText
 
-    -- * State observation and manipulation functions
-    , endOfInput
-    , atEnd
-
     -- * Utilities
     , endOfLine
+    , endOfInput
+    , match
+    , atEnd
     ) where
 
 import Control.Applicative ((<|>), (<$>))
 import Control.Monad (when)
-import Data.Attoparsec.Combinator
-import Data.Attoparsec.Internal.Types hiding (Parser, Input, Added, Failure, Success)
-import Data.Monoid (Monoid(..))
-import Data.String (IsString(..))
-import Data.Text (Text)
-import Prelude hiding (getChar, take, takeWhile)
+import Data.Attoparsec.Combinator ((<?>))
+import Data.Attoparsec.Internal
+import Data.Attoparsec.Internal.Types hiding (Parser, Failure, Success)
+import qualified Data.Attoparsec.Text.Buffer as Buf
+import Data.Attoparsec.Text.Buffer (Buffer, buffer)
 import Data.Char (chr, ord)
+import Data.List (intercalate)
+import Data.String (IsString(..))
+import Data.Text.Internal (Text(..))
+import Prelude hiding (getChar, succ, take, takeWhile)
 import qualified Data.Attoparsec.Internal.Types as T
 import qualified Data.Attoparsec.Text.FastSet as Set
 import qualified Data.Text as T
-import qualified Data.Text.Internal as T
 import qualified Data.Text.Lazy as L
+import qualified Data.Text.Unsafe as T
 
 type Parser = T.Parser Text
 type Result = IResult Text
-type Input = T.Input Text
-type Added = T.Added Text
-type Failure r = T.Failure Text r
-type Success a r = T.Success Text a r
+type Failure r = T.Failure Text Buffer r
+type Success a r = T.Success Text Buffer a r
 
 instance (a ~ Text) => IsString (Parser a) where
     fromString = string . T.pack
 
-lengthAtLeast :: T.Text -> Int -> Bool
-lengthAtLeast t@(T.Text _ _ len) n = (len `div` 2) >= n || T.length t >= n
-{-# INLINE lengthAtLeast #-}
-
--- | If at least @n@ characters of input are available, return the
--- current input, otherwise fail.
-ensure :: Int -> Parser Text
-ensure !n = T.Parser $ \i0 a0 m0 kf ks ->
-    if lengthAtLeast (unI i0) n
-    then ks i0 a0 m0 (unI i0)
-    else runParser (demandInput >> go n) i0 a0 m0 kf ks
-  where
-    go n' = T.Parser $ \i0 a0 m0 kf ks ->
-        if lengthAtLeast (unI i0) n'
-        then ks i0 a0 m0 (unI i0)
-        else runParser (demandInput >> go n') i0 a0 m0 kf ks
-{-# INLINE ensure #-}
-
--- | Ask for input.  If we receive any, pass it to a success
--- continuation, otherwise to a failure continuation.
-prompt :: Input -> Added -> More
-       -> (Input -> Added -> More -> Result r)
-       -> (Input -> Added -> More -> Result r)
-       -> Result r
-prompt i0 a0 _m0 kf ks = Partial $ \s ->
-    if T.null s
-    then kf i0 a0 Complete
-    else ks (i0 <> I s) (a0 <> A s) Incomplete
-
--- | Immediately demand more input via a 'Partial' continuation
--- result.
-demandInput :: Parser ()
-demandInput = T.Parser $ \i0 a0 m0 kf ks ->
-    if m0 == Complete
-    then kf i0 a0 m0 ["demandInput"] "not enough input"
-    else let kf' i a m = kf i a m ["demandInput"] "not enough input"
-             ks' i a m = ks i a m ()
-         in prompt i0 a0 m0 kf' ks'
-
--- | This parser always succeeds.  It returns 'True' if any input is
--- available either immediately or on demand, and 'False' if the end
--- of all input has been reached.
-wantInput :: Parser Bool
-wantInput = T.Parser $ \i0 a0 m0 _kf ks ->
-  case () of
-    _ | not (T.null (unI i0)) -> ks i0 a0 m0 True
-      | m0 == Complete  -> ks i0 a0 m0 False
-      | otherwise       -> let kf' i a m = ks i a m False
-                               ks' i a m = ks i a m True
-                           in prompt i0 a0 m0 kf' ks'
-
-get :: Parser Text
-get  = T.Parser $ \i0 a0 m0 _kf ks -> ks i0 a0 m0 (unI i0)
-
-put :: Text -> Parser ()
-put s = T.Parser $ \_i0 a0 m0 _kf ks -> ks (I s) a0 m0 ()
-
--- | Attempt a parse, and if it fails, rewind the input so that no
--- input appears to have been consumed.
---
--- This combinator is provided for compatibility with Parsec.
--- Attoparsec parsers always backtrack on failure.
-try :: Parser a -> Parser a
-try p = p
-{-# INLINE try #-}
-
-unsafeHead :: Text -> Char
-unsafeHead = T.head
-
-unsafeTail :: Text -> Text
-unsafeTail = T.tail
-
-unsafeTake :: Int -> Text -> Text
-unsafeTake = T.take
-
-unsafeDrop :: Int -> Text -> Text
-unsafeDrop = T.drop
-
 -- | The parser @satisfy p@ succeeds for any character for which the
 -- predicate @p@ returns 'True'. Returns the character that is
 -- actually parsed.
@@ -175,10 +99,10 @@ unsafeDrop = T.drop
 -- >    where isDigit c = c >= '0' && c <= '9'
 satisfy :: (Char -> Bool) -> Parser Char
 satisfy p = do
-  s <- ensure 1
-  let !w = unsafeHead s
-  if p w
-    then put (unsafeTail s) >> return w
+  (k,c) <- ensure 1
+  let !h = T.unsafeHead c
+  if p h
+    then advance k >> return h
     else fail "satisfy"
 {-# INLINE satisfy #-}
 
@@ -189,9 +113,9 @@ satisfy p = do
 -- >    where isDigit c = c >= '0' && c <= '9'
 skip :: (Char -> Bool) -> Parser ()
 skip p = do
-  s <- ensure 1
-  if p (unsafeHead s)
-    then put (unsafeTail s)
+  (k,s) <- ensure 1
+  if p (T.unsafeHead s)
+    then advance k
     else fail "skip"
 
 -- | The parser @satisfyWith f p@ transforms a character, and succeeds
@@ -199,11 +123,10 @@ skip p = do
 -- parser returns the transformed character that was parsed.
 satisfyWith :: (Char -> a) -> (a -> Bool) -> Parser a
 satisfyWith f p = do
-  s <- ensure 1
-  let c = f $! unsafeHead s
+  (k,s) <- ensure 1
+  let c = f $! T.unsafeHead s
   if p c
-    then let !t = unsafeTail s
-         in put t >> return c
+    then advance k >> return c
     else fail "satisfyWith"
 {-# INLINE satisfyWith #-}
 
@@ -211,15 +134,14 @@ satisfyWith f p = do
 -- predicate returns 'True'.
 takeWith :: Int -> (Text -> Bool) -> Parser Text
 takeWith n p = do
-  s <- ensure n
-  let (h,t) = T.splitAt n s
-  if p h
-    then put t >> return h
+  (k,s) <- ensure n
+  if p s
+    then advance k >> return s
     else fail "takeWith"
 
 -- | Consume exactly @n@ characters of input.
 take :: Int -> Parser Text
-take n = takeWith n (const True)
+take n = takeWith (max n 0) (const True)
 {-# INLINE take #-}
 
 -- | @string s@ parses a sequence of characters that identically match
@@ -235,12 +157,49 @@ take n = takeWith n (const True)
 --
 -- The reason for its failure is that the first branch is a
 -- partial match, and will consume the letters @\'f\'@ and @\'o\'@
--- before failing.  In Attoparsec, the above parser will /succeed/ on
+-- before failing.  In attoparsec, the above parser will /succeed/ on
 -- that input, because the failed first branch will consume nothing.
 string :: Text -> Parser Text
-string s = takeWith (T.length s) (==s)
+string s = string_ (stringSuspended id) id s
 {-# INLINE string #-}
 
+string_ :: (forall r. Text -> Text -> Buffer -> Pos -> More
+            -> Failure r -> Success Text r -> Result r)
+        -> (Text -> Text)
+        -> Text -> Parser Text
+string_ suspended f s0 = T.Parser $ \t pos more lose succ ->
+  let s  = f s0
+      ft = f (Buf.unbufferAt (fromPos pos) t)
+  in case T.commonPrefixes s ft of
+       Nothing
+         | T.null s          -> succ t pos more T.empty
+         | T.null ft         -> suspended s s t pos more lose succ
+         | otherwise         -> lose t pos more [] "string"
+       Just (pfx,ssfx,tsfx)
+         | T.null ssfx       -> let l = Pos (T.lengthWord16 pfx)
+                                in succ t (pos + l) more (substring pos l t)
+         | not (T.null tsfx) -> lose t pos more [] "string"
+         | otherwise         -> suspended s ssfx t pos more lose succ
+{-# INLINE string_ #-}
+
+stringSuspended :: (Text -> Text)
+                -> Text -> Text -> Buffer -> Pos -> More
+                -> Failure r
+                -> Success Text r
+                -> Result r
+stringSuspended f s000 s0 t0 pos0 more0 lose0 succ0 =
+    runParser (demandInput_ >>= go) t0 pos0 more0 lose0 succ0
+  where
+    go s' = T.Parser $ \t pos more lose succ ->
+      let s = f s'
+      in case T.commonPrefixes s0 s of
+        Nothing         -> lose t pos more [] "string"
+        Just (_pfx,ssfx,tsfx)
+          | T.null ssfx -> let l = Pos (T.lengthWord16 s000)
+                           in succ t (pos + l) more (substring pos l t)
+          | T.null tsfx -> stringSuspended f s000 ssfx t pos more lose succ
+          | otherwise   -> lose t pos more [] "string"
+
 -- | Satisfy a literal string, ignoring case.
 --
 -- Note: this function is currently quite inefficient. Unicode case
@@ -253,10 +212,9 @@ stringCI s = go 0
     go !n
       | n > T.length fs = fail "stringCI"
       | otherwise = do
-      t <- ensure n
-      let h = unsafeTake n t
-      if T.toCaseFold h == fs
-        then put (unsafeDrop n t) >> return h
+      (k,t) <- ensure n
+      if T.toCaseFold t == fs
+        then advance k >> return t
         else go (n+1)
     fs = T.toCaseFold s
 {-# INLINE stringCI #-}
@@ -264,17 +222,8 @@ stringCI s = go 0
 
 -- | Satisfy a literal string, ignoring case for characters in the ASCII range.
 asciiCI :: Text -> Parser Text
-asciiCI input = do
-  t <- ensure n
-  let h = unsafeTake n t
-  if asciiToLower h == s
-    then put (unsafeDrop n t) >> return h
-    else fail "asciiCI"
+asciiCI s = string_ (stringSuspended asciiToLower) asciiToLower s
   where
-    n = T.length input
-    s = asciiToLower input
-
-    -- convert letters in the ASCII range to lower-case
     asciiToLower = T.map f
       where
         offset = ord 'a' - ord 'A'
@@ -287,11 +236,9 @@ skipWhile :: (Char -> Bool) -> Parser ()
 skipWhile p = go
  where
   go = do
-    t <- T.dropWhile p <$> get
-    put t
-    when (T.null t) $ do
-      input <- wantInput
-      when input go
+    t <- T.takeWhile p <$> get
+    continue <- inputSpansChunks (size t)
+    when continue go
 {-# INLINE skipWhile #-}
 
 -- | Consume input as long as the predicate returns 'False'
@@ -301,8 +248,9 @@ skipWhile p = go
 -- predicate returns 'True' on the first character of input.
 --
 -- /Note/: Because this parser does not fail, do not use it with
--- combinators such as 'many', because such parsers loop until a
--- failure occurs.  Careless use will thus result in an infinite loop.
+-- combinators such as 'Control.Applicative.many', because such
+-- parsers loop until a failure occurs.  Careless use will thus result
+-- in an infinite loop.
 takeTill :: (Char -> Bool) -> Parser Text
 takeTill p = takeWhile (not . p)
 {-# INLINE takeTill #-}
@@ -314,20 +262,17 @@ takeTill p = takeWhile (not . p)
 -- predicate returns 'False' on the first character of input.
 --
 -- /Note/: Because this parser does not fail, do not use it with
--- combinators such as 'many', because such parsers loop until a
--- failure occurs.  Careless use will thus result in an infinite loop.
+-- combinators such as 'Control.Applicative.many', because such
+-- parsers loop until a failure occurs.  Careless use will thus result
+-- in an infinite loop.
 takeWhile :: (Char -> Bool) -> Parser Text
 takeWhile p = (T.concat . reverse) `fmap` go []
  where
   go acc = do
-    (h,t) <- T.span p <$> get
-    put t
-    if T.null t
-      then do
-        input <- wantInput
-        if input
-          then go (h:acc)
-          else return (h:acc)
+    h <- T.takeWhile p <$> get
+    continue <- inputSpansChunks (size h)
+    if continue
+      then go (h:acc)
       else return (h:acc)
 
 takeRest :: Parser [Text]
@@ -338,7 +283,7 @@ takeRest = go []
     if input
       then do
         s <- get
-        put T.empty
+        advance (size s)
         go (s:acc)
       else return (reverse acc)
 
@@ -351,25 +296,10 @@ takeLazyText :: Parser L.Text
 takeLazyText = L.fromChunks `fmap` takeRest
 
 data Scan s = Continue s
-            | Finished {-# UNPACK #-} !Int T.Text
+            | Finished {-# UNPACK #-} !Int Text
 
--- | A stateful scanner.  The predicate consumes and transforms a
--- state argument, and each transformed state is passed to successive
--- invocations of the predicate on each character of the input until one
--- returns 'Nothing' or the input ends.
---
--- This parser does not fail.  It will return an empty string if the
--- predicate returns 'Nothing' on the first character of input.
---
--- /Note/: Because this parser does not fail, do not use it with
--- combinators such as 'many', because such parsers loop until a
--- failure occurs.  Careless use will thus result in an infinite loop.
-scan :: s -> (s -> Char -> Maybe s) -> Parser Text
-scan s0 p = do
-  chunks <- go [] s0
-  case chunks of
-    [x] -> return x
-    xs  -> return . T.concat . reverse $ xs
+scan_ :: (s -> [Text] -> Parser r) -> s -> (s -> Char -> Maybe s) -> Parser r
+scan_ f s0 p = go [] s0
  where
   scanner s !n t =
     case T.uncons t of
@@ -380,14 +310,39 @@ scan s0 p = do
   go acc s = do
     input <- get
     case scanner s 0 input of
-      Continue s'  -> do put T.empty
-                         more <- wantInput
-                         if more
+      Continue s'  -> do continue <- inputSpansChunks (size input)
+                         if continue
                            then go (input : acc) s'
-                           else return (input : acc)
-      Finished n t -> put t >> return (T.take n input : acc)
+                           else f s' (input : acc)
+      Finished n t -> do advance (size input - size t)
+                         f s (T.take n input : acc)
+{-# INLINE scan_ #-}
+
+-- | A stateful scanner.  The predicate consumes and transforms a
+-- state argument, and each transformed state is passed to successive
+-- invocations of the predicate on each character of the input until one
+-- returns 'Nothing' or the input ends.
+--
+-- This parser does not fail.  It will return an empty string if the
+-- predicate returns 'Nothing' on the first character of input.
+--
+-- /Note/: Because this parser does not fail, do not use it with
+-- combinators such as 'Control.Applicative.many', because such
+-- parsers loop until a failure occurs.  Careless use will thus result
+-- in an infinite loop.
+scan :: s -> (s -> Char -> Maybe s) -> Parser Text
+scan = scan_ $ \_ chunks ->
+  case chunks of
+    [x] -> return x
+    xs  -> return . T.concat . reverse $ xs
 {-# INLINE scan #-}
 
+-- | Like 'scan', but generalized to return the final state of the
+-- scanner.
+runScanner :: s -> (s -> Char -> Maybe s) -> Parser (Text, s)
+runScanner = scan_ $ \s xs -> return (T.concat (reverse xs), s)
+{-# INLINE runScanner #-}
+
 -- | Consume input as long as the predicate returns 'True', and return
 -- the consumed input.
 --
@@ -396,11 +351,13 @@ scan s0 p = do
 -- 'True' or if there is no input left.
 takeWhile1 :: (Char -> Bool) -> Parser Text
 takeWhile1 p = do
-  (`when` demandInput) =<< T.null <$> get
-  (h,t) <- T.span p <$> get
-  when (T.null h) $ fail "takeWhile1"
-  put t
-  if T.null t
+  (`when` demandInput) =<< endOfChunk
+  h <- T.takeWhile p <$> get
+  let size' = size h
+  when (size' == 0) $ fail "takeWhile1"
+  advance size'
+  eoc <- endOfChunk
+  if eoc
     then (h<>) `fmap` takeWhile p
     else return h
 
@@ -440,79 +397,143 @@ notChar :: Char -> Parser Char
 notChar c = satisfy (/= c) <?> "not " ++ show c
 {-# INLINE notChar #-}
 
--- | Match any character. Returns 'Nothing' if end of input has been
--- reached. Does not consume any input.
+-- | Match any character, to perform lookahead. Returns 'Nothing' if
+-- end of input has been reached. Does not consume any input.
 --
 -- /Note/: Because this parser does not fail, do not use it with
--- combinators such as 'many', because such parsers loop until a
--- failure occurs.  Careless use will thus result in an infinite loop.
+-- combinators such as 'Control.Applicative.many', because such
+-- parsers loop until a failure occurs.  Careless use will thus result
+-- in an infinite loop.
 peekChar :: Parser (Maybe Char)
-peekChar = T.Parser $ \i0 a0 m0 _kf ks ->
-           if T.null (unI i0)
-           then if m0 == Complete
-                then ks i0 a0 m0 Nothing
-                else let ks' i a m = let !c = unsafeHead (unI i)
-                                     in ks i a m (Just c)
-                         kf' i a m = ks i a m Nothing
-                     in prompt i0 a0 m0 kf' ks'
-           else let !c = unsafeHead (unI i0)
-                in ks i0 a0 m0 (Just c)
+peekChar = T.Parser $ \t pos more _lose succ ->
+  case () of
+    _| pos < lengthOf t ->
+       let T.Iter !c _ = Buf.iter t (fromPos pos)
+       in succ t pos more (Just c)
+     | more == Complete ->
+       succ t pos more Nothing
+     | otherwise ->
+       let succ' t' pos' more' =
+             let T.Iter !c _ = Buf.iter t' (fromPos pos')
+             in succ t' pos' more' (Just c)
+           lose' t' pos' more' = succ t' pos' more' Nothing
+       in prompt t pos more lose' succ'
 {-# INLINE peekChar #-}
 
--- | Match only if all input has been consumed.
-endOfInput :: Parser ()
-endOfInput = T.Parser $ \i0 a0 m0 kf ks ->
-             if T.null (unI i0)
-             then if m0 == Complete
-                  then ks i0 a0 m0 ()
-                  else let kf' i1 a1 m1 _ _ = addS i0 a0 m0 i1 a1 m1 $
-                                              \ i2 a2 m2 -> ks i2 a2 m2 ()
-                           ks' i1 a1 m1 _   = addS i0 a0 m0 i1 a1 m1 $
-                                              \ i2 a2 m2 -> kf i2 a2 m2 []
-                                                            "endOfInput"
-                       in  runParser demandInput i0 a0 m0 kf' ks'
-             else kf i0 a0 m0 [] "endOfInput"
-
--- | Return an indication of whether the end of input has been
--- reached.
-atEnd :: Parser Bool
-atEnd = not <$> wantInput
-{-# INLINE atEnd #-}
+-- | Match any character, to perform lookahead.  Does not consume any
+-- input, but will fail if end of input has been reached.
+peekChar' :: Parser Char
+peekChar' = do
+  (_,s) <- ensure 1
+  return $! T.unsafeHead s
+{-# INLINE peekChar' #-}
 
 -- | Match either a single newline character @\'\\n\'@, or a carriage
 -- return followed by a newline character @\"\\r\\n\"@.
 endOfLine :: Parser ()
 endOfLine = (char '\n' >> return ()) <|> (string "\r\n" >> return ())
 
--- | Name the parser, in case failure occurs.
-(<?>) :: Parser a
-      -> String                 -- ^ the name to use if parsing fails
-      -> Parser a
-p <?> msg0 = T.Parser $ \i0 a0 m0 kf ks ->
-             let kf' i a m strs msg = kf i a m (msg0:strs) msg
-             in runParser p i0 a0 m0 kf' ks
-{-# INLINE (<?>) #-}
-infix 0 <?>
-
 -- | Terminal failure continuation.
 failK :: Failure a
-failK i0 _a0 _m0 stack msg = Fail (unI i0) stack msg
+failK t (Pos pos) _more stack msg = Fail (Buf.dropWord16 pos t) stack msg
 {-# INLINE failK #-}
 
 -- | Terminal success continuation.
 successK :: Success a a
-successK i0 _a0 _m0 a = Done (unI i0) a
+successK t (Pos pos) _more a = Done (Buf.dropWord16 pos t) a
 {-# INLINE successK #-}
 
 -- | Run a parser.
 parse :: Parser a -> Text -> Result a
-parse m s = runParser m (I s) mempty Incomplete failK successK
+parse m s = runParser m (buffer s) 0 Incomplete failK successK
 {-# INLINE parse #-}
 
 -- | Run a parser that cannot be resupplied via a 'Partial' result.
+--
+-- This function does not force a parser to consume all of its input.
+-- Instead, any residual input will be discarded.  To force a parser
+-- to consume all of its input, use something like this:
+--
+-- @
+--'parseOnly' (myParser 'Control.Applicative.<*' 'endOfInput')
+-- @
 parseOnly :: Parser a -> Text -> Either String a
-parseOnly m s = case runParser m (I s) mempty Complete failK successK of
-                  Fail _ _ err -> Left err
-                  Done _ a     -> Right a
-                  _            -> error "parseOnly: impossible error!"
+parseOnly m s = case runParser m (buffer s) 0 Complete failK successK of
+                  Fail _ [] err   -> Left err
+                  Fail _ ctxs err -> Left (intercalate " > " ctxs ++ ": " ++ err)
+                  Done _ a        -> Right a
+                  _               -> error "parseOnly: impossible error!"
 {-# INLINE parseOnly #-}
+
+get :: Parser Text
+get = T.Parser $ \t pos more _lose succ ->
+  succ t pos more (Buf.dropWord16 (fromPos pos) t)
+{-# INLINE get #-}
+
+endOfChunk :: Parser Bool
+endOfChunk = T.Parser $ \t pos more _lose succ ->
+  succ t pos more (pos == lengthOf t)
+{-# INLINE endOfChunk #-}
+
+inputSpansChunks :: Pos -> Parser Bool
+inputSpansChunks i = T.Parser $ \t pos_ more _lose succ ->
+  let pos = pos_ + i
+  in if pos < lengthOf t || more == Complete
+     then succ t pos more False
+     else let lose' t' pos' more' = succ t' pos' more' False
+              succ' t' pos' more' = succ t' pos' more' True
+          in prompt t pos more lose' succ'
+{-# INLINE inputSpansChunks #-}
+
+advance :: Pos -> Parser ()
+advance n = T.Parser $ \t pos more _lose succ -> succ t (pos+n) more ()
+{-# INLINE advance #-}
+
+ensureSuspended :: Int -> Buffer -> Pos -> More
+                -> Failure r -> Success (Pos, Text) r
+                -> Result r
+ensureSuspended n t pos more lose succ =
+    runParser (demandInput >> go) t pos more lose succ
+  where go = T.Parser $ \t' pos' more' lose' succ' ->
+          case lengthAtLeast pos' n t' of
+            Just n' -> succ' t' pos' more' (n', substring pos n' t')
+            Nothing -> runParser (demandInput >> go) t' pos' more' lose' succ'
+
+-- | If at least @n@ elements of input are available, return the
+-- current input, otherwise fail.
+ensure :: Int -> Parser (Pos, Text)
+ensure n = T.Parser $ \t pos more lose succ ->
+    case lengthAtLeast pos n t of
+      Just n' -> succ t pos more (n', substring pos n' t)
+      -- The uncommon case is kept out-of-line to reduce code size:
+      Nothing -> ensureSuspended n t pos more lose succ
+{-# INLINE ensure #-}
+
+-- | Return both the result of a parse and the portion of the input
+-- that was consumed while it was being parsed.
+match :: Parser a -> Parser (Text, a)
+match p = T.Parser $ \t pos more lose succ ->
+  let succ' t' pos' more' a = succ t' pos' more'
+                              (substring pos (pos'-pos) t', a)
+  in runParser p t pos more lose succ'
+
+-- | Ensure that at least @n@ code points of input are available.
+-- Returns the number of words consumed while traversing.
+lengthAtLeast :: Pos -> Int -> Buffer -> Maybe Pos
+lengthAtLeast pos n t = go 0 (fromPos pos)
+  where go i !p
+          | i == n    = Just (Pos p - pos)
+          | p == len  = Nothing
+          | otherwise = go (i+1) (p + Buf.iter_ t p)
+        Pos len = lengthOf t
+{-# INLINE lengthAtLeast #-}
+
+substring :: Pos -> Pos -> Buffer -> Text
+substring (Pos pos) (Pos n) = Buf.substring pos n
+{-# INLINE substring #-}
+
+lengthOf :: Buffer -> Pos
+lengthOf = Pos . Buf.length
+
+size :: Text -> Pos
+size (Text _ _ l) = Pos l
